/*******************************************************************************
 *  RCOM : COM Client and Server for R
 *  Copyright (C) 2003-2005 Thomas Baier
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  ---------------------------------------------------------------------------
 *
 *  $Id: public.cpp 1.1 2003/04/27 16:32:04 baier Exp $
 *
 ******************************************************************************/

#include "rcom.h"
#include <stdio.h>
#include "Rinternals.h"
#include "bdx_util.h"
#include "bdx_com.h"

/* SAMPLE SESSION WITH MICROSOFT EXCEL

library(rcom);
excel<-comCreate("Excel.Application");
comPropertySet(excel,"Visible",TRUE);
wb<-comInvoke(comPropertyGet(excel,"Workbooks"),"Open","c:\\tmp\\RTS Beteiligung.xls");
sheet<-comPropertyGet(comPropertyGet(wb,"Sheets"),"Item","Kalkulationen");
comInvoke(sheet,"Activate");
r1<-comPropertyGet(sheet,"Range","C5");
r2<-comPropertyGet(sheet,"Range","C6","C8");
r3<-comPropertyGet(sheet,"Range","C6","E8");
r4<-comPropertyGet(sheet,"Range","C3");
r5<-comPropertyGet(sheet,"Range","C3","E3");
 */

/* sample session with power point

ppt<-com.object.create("PowerPoint.Application")
presentations<-comPropertyGet(ppt,"Presentations")
comInvoke(presentations,"Add")
# take care: Item property is not a property!!!
pres<-comInvoke(presentations,"Item",1)
slides<-comPropertyGet(pres,"Slides")
# type 12 is empty slide, 8 is chart, 2 is title+enumeration
slide<-comInvoke(slides,"Add",2,1)
# set the title of the slide to "Summary Information"
title<-comPropertyGet(comPropertyGet(slide,"Shapes"),"Title")
textrange<-comPropertyGet(comPropertyGet(title,"TextFrame"),"TextRange")
comPropertySet(textrange,"Text","Summary Information")
# add analysis results to slide
enumshape<-comInvoke(comPropertyGet(slide,"Shapes"),"Item",2)
textrange<-comPropertyGet(comPropertyGet(enumshape,"TextFrame"),"TextRange")
comPropertySet(textrange,"Text","Text1\rText2\rTexgt3")

 */
extern "C" {

static int isInitialized = 0;

void com_initialize(void)
{
  if (!isInitialized) {
    CoInitialize (NULL);
    isInitialized = 1;
  }
}

/** Create a COM Object.
 *
 * Call #com_release()# when you're done with the object.
 *
 * @arg progid[0] is the program id of the COM object to create,
 *      e.g. #"Excel.Application"#
 * @return rc[0] = 0 on error, else the object handle
 */
SEXP com_create(SEXP progid)
{
  CLSID clsid;
  HRESULT hr;
  LPDISPATCH object = NULL;
  SEXP sexp = NULL;
  RCOM_OBJHANDLE handle = RCOM_NULLHANDLE;

  if (TYPEOF(progid) != STRSXP) {
    RCOM_ERR(printf ("com_create(): progid not a string\n"));
    return R_NilValue;
  }
  com_initialize();
  hr = com_getCLSID (CHAR (STRING_ELT(progid,0)),&clsid);
  if (FAILED (hr)) {
    RCOM_ERR(printf ("com_create: failed error %08lx querying CLSID for \"%s\"\n",
		   hr,CHAR (STRING_ELT(progid,0))));
    return R_NilValue;
  }

  hr = CoCreateInstance (clsid,NULL,CLSCTX_SERVER,
			 IID_IDispatch,(LPVOID*) &object);

  if (FAILED (hr)) {
    RCOM_ERR(printf ("com_create: failed error %08lx creating \"%s\"\n",
		   hr,CHAR (STRING_ELT(progid,0))));
    return R_NilValue;
  }

  handle = com_addObject (object);
  sexp = com_createSEXP(handle);
  if(sexp == R_NilValue) {
    RCOM_ERR(printf("com_create: invalid SEXP for object, releasing object\n"));
    object->Release();
  }
  return sexp;
}


/** Create a COM Object.
 *
 * Call #com_release()# when you're done with the object.
 *
 * @arg progid[0] is the program id of the COM object to create,
 *      e.g. #"Excel.Application"#
 * @return rc[0] = 0 on error, else the object handle
 */
SEXP com_getobject(SEXP progid)
{
  CLSID clsid;
  HRESULT hr;
  LPDISPATCH object = NULL;
  SEXP sexp = NULL;
  RCOM_OBJHANDLE handle = RCOM_NULLHANDLE;

  if (TYPEOF(progid) != STRSXP) {
    RCOM_ERR(printf ("com_getobject(): progid not a string\n"));
    return R_NilValue;
  }
  com_initialize();
  hr = com_getCLSID (CHAR (STRING_ELT(progid,0)),&clsid);
  if (FAILED (hr)) {
    RCOM_ERR(printf ("com_getobject: failed error %08lx querying CLSID for \"%s\"\n",
		   hr,CHAR (STRING_ELT(progid,0))));
    return R_NilValue;
  }
  {
    IUnknown* unk;
    hr = GetActiveObject (clsid,NULL,&unk);
    if (FAILED (hr)) {
      RCOM_ERR(printf ("com_getobject: failed error %08lx getting \"%s\"\n",
		       hr,CHAR (STRING_ELT(progid,0))));
      return R_NilValue;
    }
    hr = unk->QueryInterface(IID_IDispatch,(LPVOID*) &object);
    if (FAILED (hr)) {
      RCOM_ERR(printf ("com_getobject: failed error %08lx querying IDispatch\n",
		       hr));
      unk->Release();
      return R_NilValue;
    }
    unk->Release();
  }
  handle = com_addObject (object);
  sexp = com_createSEXP(handle);
  if(sexp == R_NilValue) {
    RCOM_ERR(printf("com_create: invalid SEXP for object, releasing object\n"));
    object->Release();
  }
  return sexp;
}


/** Check for a valid COM Object.
 *
 * @arg handle[0] is the object handle as returned by com_create().
 * @return rc[0] is true if handle[0] is a valid object, false
 * otherwise
 */
SEXP com_isvalid(SEXP handle)
{
  SEXP rc = PROTECT (allocVector (LGLSXP,1));
  if (com_getObject (com_getHandle(handle)) != RCOM_NULLHANDLE) {
    LOGICAL(rc)[0] = 1;
  } else {
    LOGICAL(rc)[0] = 0;
  }
  UNPROTECT(1);
  return rc;
}

/** Set a Property.
 *
 * @arg handle is the object handle as returned by com_create().
 * @arg property is the name of the property
 * @arg value is the new property value
 */
SEXP com_property_set(SEXP handle,SEXP property,SEXP value)
{
  VARIANTARG _var;
  DISPID propid = 0;
  DISPID dspput = DISPID_PROPERTYPUT;
  HRESULT hr;
  LPDISPATCH dsp;
  DISPPARAMS _value = {
    &_var,
    &dspput,
    1,
    1
  };

  dsp = com_getObject (com_getHandle (handle));
  if (dsp == NULL) {
    RCOM_ERR(printf ("com_property_set: handle not valid\n"));
    return R_NilValue;
  }
  if (TYPEOF (property) != STRSXP) {
    RCOM_ERR(printf ("com_property_set: property name not a string\n"));
    return R_NilValue;
  }
  hr = com_GetIDsOfName (dsp,CHAR(STRING_ELT(property,0)),&propid);
  if (FAILED(hr)) {
    RCOM_ERR(printf ("com_property_set: property \"%s\" not found (%08lx)\n",
		     CHAR(STRING_ELT(property,0)),hr));
    return R_NilValue;
  }
  /*  if (LENGTH (value) != 1) {
    RCOM_TRACE(printf ("com_property_set: only scalars and strings supported\n"));
    return R_NilValue;
    }*/
  VariantInit(&_var);
  if (!SEXP2Variant (value,&_var)) {
    RCOM_ERR(printf ("com_property_set: conversion to variant failed\n"));
    return R_NilValue;
  }
  hr = dsp->Invoke(propid,IID_NULL,LOCALE_USER_DEFAULT,DISPATCH_PROPERTYPUT,
			   &_value,NULL,NULL,NULL);
  if (FAILED (hr)) {
      RCOM_ERR(printf ("com_property_set: error %08lx setting property \"%s\"\n",
		       hr,CHAR(STRING_ELT(property,0))));
      return R_NilValue;
  }
  return R_NilValue;
}

/** Get a Property.
 *
 * @arg handle is the object handle as returned by com_create().
 * @arg property is the name of the property
 * @arg index is the name/index for a collection element (e.g. Item property)
 *
 * 05-06-13 | baier | reversed order for anonymous parameters 
 */
SEXP com_property_get(SEXP handle,SEXP property,SEXP index)
{
  VARIANT _result;
  DISPID propid = 0;
  VARIANTARG* _var = NULL;
  HRESULT hr;
  LPDISPATCH dsp;
  DISPPARAMS _value = {
    NULL,
    NULL,
    0,
    0
  };
  SEXP result = NULL;

  dsp = com_getObject (com_getHandle (handle));
  RCOM_TRACE(printf("com_property_get(%p,\"%s\")\n",
		    dsp,CHAR(STRING_ELT(property,0))));
  if (dsp == NULL) {
    RCOM_ERR(printf ("com_property_get: handle not valid\n"));
    return R_NilValue;
  }
  if (TYPEOF (property) != STRSXP) {
    RCOM_ERR(printf ("com_property_get: property name not a string\n"));
    return R_NilValue;
  }
  hr = com_GetIDsOfName (dsp,CHAR(STRING_ELT(property,0)),&propid);
  if (FAILED(hr)) {
    RCOM_ERR(printf ("com_property_get: property \"%s\" not found (%08lx)\n",
		     CHAR(STRING_ELT(property,0)),hr));
    return R_NilValue;
  }
  int argc = 0;
  if (index == R_NilValue) {
    RCOM_TRACE(printf("com_property_get: no parameters\n"));
    argc = 0;
  } else if (TYPEOF(index) == VECSXP) {
    RCOM_TRACE(printf ("com_property_get: list of arguments, counting parameters\n"));
    argc = LENGTH(index);
    RCOM_TRACE(printf("com_property_get: %d parameters for function counted\n",
		 argc));
  } else {
    RCOM_TRACE(printf("com_property_get: single parameter, type %d\n",TYPEOF(index)));
    com_trace(index);
    argc = 1;
  }
  if (argc == 0) {
    _var = NULL;
  } else {
    _var = (VARIANTARG*) malloc(sizeof(VARIANTARG) * argc);
    for (int i = 0;i < argc;i++) {
      VariantInit(&_var[argc-i-1]);
      RCOM_TRACE(printf("Converting argument %d to VARIANT\n",i));
      if (!SEXP2Variant(VECTOR_ELT(index,i),&_var[argc-i-1])) {
	RCOM_ERR(printf("com_property_get: conversion of parameter %d to variant failed\n",
			i));
	free(_var);
	return R_NilValue;
      }
    }
  }
  _value.rgvarg = _var;
  _value.cArgs = argc;
  VariantInit(&_result);
  hr = dsp->Invoke(propid,IID_NULL,LOCALE_USER_DEFAULT,DISPATCH_PROPERTYGET,
		   &_value,&_result,NULL,NULL);
  if (argc > 0) free(_var);
  if (FAILED (hr)) {
      RCOM_ERR(printf ("com_property_get: error %08lx getting property \"%s\"\n",
		       hr,CHAR(STRING_ELT(property,0))));
      return R_NilValue;
  }
  if (!Variant2SEXP (&_result,&result)) {
    RCOM_ERR(printf ("com_property_get: conversion to SEXP failed\n"));
    return R_NilValue;
  }
  VariantClear (&_result);
  return result;
}

/** Call a COM Method.
 *
 * @arg handle is the object handle as returned by com_create().
 * @arg method is the name of the method to invoke
 * @arg property is the name of the property
 *
 * 04-07-04 | baier | better information on errors
 * 04-07-05 | baier | positional arguments in reversed order!
 */
SEXP com_invoke(SEXP handle,SEXP method,SEXP args)
{
  VARIANT _result;
  VARIANTARG* _var = NULL;
  DISPID methodid = 0;
  HRESULT hr;
  LPDISPATCH dsp;
  DISPPARAMS _value = {
    NULL,
    NULL, // &dspput,
    0, //1,
    0 // 1
  };
  SEXP result = NULL;
  unsigned int argErr;

  dsp = com_getObject (com_getHandle (handle));
  if (dsp == NULL) {
    RCOM_ERR(printf ("com_invoke: handle not valid\n"));
    return R_NilValue;
  }
  if (TYPEOF (method) != STRSXP) {
    RCOM_ERR(printf ("com_invoke: property name not a string\n"));
    return R_NilValue;
  }
  hr = com_GetIDsOfName (dsp,CHAR(STRING_ELT(method,0)),&methodid);
  if (FAILED(hr)) {
    RCOM_ERR(printf ("com_invoke: method \"%s\" not found (%08lx)\n",
		 CHAR(STRING_ELT(method,0)),hr));
    return R_NilValue;
  }
  // prepare parameters: count and allocate variant data for all parameters
  int argc = 0;
  if (args == R_NilValue) {
    RCOM_TRACE(printf("com_invoke: no parameters\n"));
    argc = 0;
  } else if (TYPEOF(args) == VECSXP) {
    RCOM_TRACE(printf ("com_invoke: list of arguments, counting parameters\n"));
    argc = LENGTH(args);
    RCOM_TRACE(printf("com_invoke: %d parameters for function counted\n",
		 argc));
  } else {
    RCOM_TRACE(printf("com_invoke: single parameter, type %d\n",TYPEOF(args)));
    com_trace(args);
    argc = 1;
  }
  if (argc == 0) {
    _var = NULL;
  } else {
    _var = (VARIANTARG*) malloc(sizeof(VARIANTARG) * argc);
    for (int i = 0;i < argc;i++) {
      VariantInit(&(_var[argc-i-1]));
      RCOM_TRACE(printf("Converting argument %d to VARIANT\n",i));
      if (!SEXP2Variant(VECTOR_ELT(args,i),&(_var[argc-i-1]))) {
	RCOM_ERR(printf("com_invoke: conversion of parameter %d to variant failed\n",
			i));
	free(_var);
	return R_NilValue;
      }
    }
  }
  _value.rgvarg = _var;
  _value.cArgs = argc;
  VariantInit(&_result);
  hr = dsp->Invoke(methodid,IID_NULL,LOCALE_USER_DEFAULT,DISPATCH_METHOD,
		   &_value,&_result,NULL,&argErr);
  if (argc > 0) free(_var);
  if (FAILED (hr)) {
    switch(hr) {
    case DISP_E_TYPEMISMATCH:
      RCOM_ERR(printf ("com_invoke: \"type mismatch\" for parameter #%d invoking method \"%s\"\n",
		       argErr,CHAR(STRING_ELT(method,0))));
      break;
    case DISP_E_BADPARAMCOUNT:
      RCOM_ERR(printf("com_invoke: \"bad parameter count\" invoking method \"%s\"\n",
		      CHAR(STRING_ELT(method,0))));
      break;
    case DISP_E_BADVARTYPE:
      RCOM_ERR(printf("com_invoke: \"bad variant type\" invoking method \"%s\"\n",
		      CHAR(STRING_ELT(method,0))));
      break;
    case DISP_E_PARAMNOTFOUND:
      RCOM_ERR(printf("com_invoke: \"parameter not found\" for parameter #%d invoking method \"%s\"\n",
		      argErr,CHAR(STRING_ELT(method,0))));
      break;
    case DISP_E_OVERFLOW:
      RCOM_ERR(printf("com_invoke: \"overflow\" invoking method \"%s\"\n",
		      CHAR(STRING_ELT(method,0))));
      break;
    default:
      RCOM_ERR(printf ("com_invoke: error %08lx invoking method \"%s\"\n",
		       hr,CHAR(STRING_ELT(method,0))));
    }
    return R_NilValue;
  }
  if (!Variant2SEXP (&_result,&result)) {
    RCOM_ERR(printf ("com_invoke: conversion of result to SEXP failed\n"));
    return R_NilValue;
  }
  VariantClear (&_result);
  return result;
}

HRESULT com_GetIDsOfName(LPDISPATCH dsp,char* name,DISPID* dispid)
{
  HRESULT hr;
  OLECHAR* _name = com_getOLECHAR (name);
  if (_name == NULL) {
    RCOM_ERR(printf ("com_GetIDsOfName: \"%s\" not a valid string\n",name));
    return E_FAIL;
  }
  hr = dsp->GetIDsOfNames(IID_NULL,&_name,1,LOCALE_USER_DEFAULT,dispid);
  free (_name); _name = NULL;
  return hr;
}


void com_trace_object(int level,LPDISPATCH obj)
{
  ITypeInfo* typeinfo = NULL;
  HRESULT hr = E_FAIL;

  hr = obj->GetTypeInfo(0,LOCALE_USER_DEFAULT,&typeinfo);
  if(FAILED(hr)) {
    RCOM_ERR(printf("com_trace_object(): GetTypeInfo() returns failure code %08lx\n",
		    hr));
    return;
  }
  typeinfo->Release();
  /* properties first */
}


SEXP com_trace_internal (int level,SEXP sexp)
{
  char* structtype = NULL;
  int length = 0;
  int dims = 0;

  if(sexp == R_NilValue) {
    RCOM_ERR(printf("(%d) SEXP is R_NilValue\n",level));
    return R_NilValue;
  }
  RCOM_ERR(printf("(%d) SEXP is %p\n",level,sexp));

  if (LENGTH (sexp) == 0) {
    RCOM_ERR(printf ("(%d) SEXP has length 0 -- invalid\n",level));
    return R_NilValue;
  } else if (LENGTH (sexp) == 1) {
    structtype = "scalar";
    length = 1;
    dims = 1;
  } else {
    RCOM_ERR(printf("(%d) checking SEXP, length is %d, type %d\n",
		    level,LENGTH(sexp),TYPEOF(sexp)));
    SEXP dim = getAttrib (sexp,R_DimSymbol);
    length = LENGTH (sexp);
    if (TYPEOF (dim) == NILSXP) {
      RCOM_ERR(printf("dim is NILSXP"));
      structtype = "vector";
      dims = LENGTH (sexp);
    } else {
      RCOM_ERR(printf("dim is not NILSXP"));
      structtype = "matrix";
      dims = INTEGER (dim)[0];
    }
  }
  RCOM_ERR(printf ("(%d) SEXP is a %s, %d elements, %d dimensions\n",level,structtype,
		   length,dims));
  switch (TYPEOF(sexp)) {
  case NILSXP:
    RCOM_ERR(printf ("(%d) SEXP is of type NILSXP\n",level));
    return R_NilValue;
  case LGLSXP:
    RCOM_ERR(printf ("(%d) SEXP is of type LGLSXP, value %s\n",level,
		     LOGICAL (sexp)[0] ? "TRUE" : "FALSE"));
    break;
  case INTSXP:
    RCOM_ERR(printf ("(%d) SEXP is of type INTSXP, value %d\n",level,INTEGER (sexp)[0]));
    break;
  case REALSXP:
    RCOM_ERR(printf ("(%d) SEXP is of type REALSXP, value %f\n",level,REAL (sexp)[0]));
    break;
  case STRSXP:
    RCOM_ERR(printf ("(%d) SEXP is of type STRSXP, value \"%s\"\n",level,CHAR(STRING_ELT(sexp,0))));
    break;
  case VECSXP:
    {
      int i;
      RCOM_ERR(printf ("(%d) SEXP is of type VECSXP, now tracing contents\n",level));
      for (i = 0;i < LENGTH (sexp);i++) {
	com_trace_internal (level+1,VECTOR_ELT (sexp,i));
      }
    }
    break;
  case EXTPTRSXP:
    {
      // is it a COM object?
      RCOM_OBJHANDLE handle = com_getHandle (sexp);
      if (handle != RCOM_NULLHANDLE) {
	// valid COM object
	RCOM_ERR(printf ("(%d) SEXP is of type EXTPTRSXP, is a valid COM handle (%p, %p)\n",level,
			   handle,com_getObject (handle)));
	com_trace_object(level,com_getObject(handle));
      } else {
	// plain external pointer
	RCOM_ERR(printf ("(%d) SEXP is of type EXTPTRSXP, value %p\n",
			   level,R_ExternalPtrAddr(sexp)));
      }
    }
    break;
  case CPLXSXP:
    RCOM_ERR(printf ("(%d) SEXP is of type CPLXSXP\n",level));
    break;
  default:
    RCOM_ERR(printf ("(%d) SEXP is of unknown type %d\n",level,TYPEOF (sexp)));
    return R_NilValue;
  }
  return R_NilValue;
}

SEXP com_trace(SEXP sexp)
{
  return com_trace_internal (0,sexp);
}


void my_finalizer(SEXP sexp)
{
  if(TYPEOF(sexp) != EXTPTRSXP) {
    RCOM_ERR(printf("my_finalizer() called for non-pointer SEXP\n"));
  } else {
    RCOM_TRACE(printf("my_finalizer() called for pointer at %p\n",
		      R_ExternalPtrAddr(sexp)));
  }
}

char* get_elem_desc(ELEMDESC* desc)
{
  char tmp[1024];
  char tmp2[1024];
  switch(desc->tdesc.vt) {
  case VT_EMPTY: sprintf(tmp,"VT_EMPTY"); break;
  case VT_NULL: sprintf(tmp,"VT_NULL"); break;
  case VT_I2: sprintf(tmp,"short"); break;
  case VT_I4: sprintf(tmp,"int"); break;
  case VT_R4: sprintf(tmp,"real"); break;
  case VT_R8: sprintf(tmp,"double"); break;
  case VT_CY: sprintf(tmp,"CURRENCY"); break;
  case VT_DATE: sprintf(tmp,"DATE"); break;
  case VT_BSTR: sprintf(tmp,"BSTR"); break;
  case VT_DISPATCH: sprintf(tmp,"LPDISPATCH"); break;
  case VT_ERROR: sprintf(tmp,"VT_ERROR"); break;
  case VT_BOOL: sprintf(tmp,"VARIANT_BOOL"); break;
  case VT_VARIANT: sprintf(tmp,"VARIANT"); break;
  case VT_DECIMAL: sprintf(tmp,"DECIMAL"); break;
  case VT_RECORD: sprintf(tmp,"VT_RECORD"); break;
  case VT_UNKNOWN: sprintf(tmp,"VT_UNKNOWN"); break;
  case VT_I1: sprintf(tmp,"char"); break;
  case VT_UI1: sprintf(tmp,"unsigned char"); break;
  case VT_UI2: sprintf(tmp,"unsigned short"); break;
  case VT_UI4: sprintf(tmp,"unsigned long"); break;
  case VT_INT: sprintf(tmp,"int"); break;
  case VT_UINT: sprintf(tmp,"unsigned int"); break;
  case VT_VOID: sprintf(tmp,"void"); break;
  case VT_HRESULT: sprintf(tmp,"HRESULT"); break;
  case VT_PTR: sprintf(tmp,"POINTER"); break;
  case VT_SAFEARRAY: sprintf(tmp,"SAFEARRAY"); break;
  case VT_USERDEFINED: sprintf(tmp,"VT_USERDEFINED"); break;
  case VT_LPSTR: sprintf(tmp,"LPSTR"); break;
  case VT_LPWSTR: sprintf(tmp,"LPWSTR"); break;
  case VT_FILETIME: sprintf(tmp,"FILETIME"); break;
  case VT_BLOB: sprintf(tmp,"BLOB"); break;
  case VT_STREAM: sprintf(tmp,"STREAM"); break;
  case VT_STORAGE: sprintf(tmp,"STORAGE"); break;
  case VT_STREAMED_OBJECT: sprintf(tmp,"STREAMED_OBJECT"); break;
  case VT_STORED_OBJECT: sprintf(tmp,"STORED_OBJECT"); break;
  case VT_BLOB_OBJECT: sprintf(tmp,"BLOB_OBJECT"); break;
  case VT_CF: sprintf(tmp,"CLIPFORMAT"); break;
  case VT_CLSID: sprintf(tmp,"CLSID"); break;
  case VT_VECTOR: sprintf(tmp,"VECTOR"); break;
  case VT_ARRAY: sprintf(tmp,"ARRAY"); break;
  case VT_BYREF: sprintf(tmp,"VT_BYREF"); break;
  case VT_RESERVED: sprintf(tmp,"VT_RESERVED"); break;
  default: sprintf(tmp,"unknown type"); break;
  }
  char params[1024];
  strcpy(params,"");
  if(desc->paramdesc.wParamFlags != PARAMFLAG_NONE) {
    int first = 1;
    if((desc->paramdesc.wParamFlags & PARAMFLAG_FIN) == PARAMFLAG_FIN) {
      if(first) {
	sprintf(params,"in");
	first = 0;
      } else {
	sprintf(params,"%s,in",params);
      }
    }
    if((desc->paramdesc.wParamFlags & PARAMFLAG_FOUT) == PARAMFLAG_FOUT) {
      if(first) {
	sprintf(params,"out");
	first = 0;
      } else {
	sprintf(params,"%s,out",params);
      }
    }
    if((desc->paramdesc.wParamFlags & PARAMFLAG_FRETVAL) == PARAMFLAG_FRETVAL) {
      if(first) {
	sprintf(params,"retval");
	first = 0;
      } else {
	sprintf(params,"%s,retval",params);
      }
    }
    if((desc->paramdesc.wParamFlags & PARAMFLAG_FOPT) == PARAMFLAG_FOPT) {
      if(first) {
	sprintf(params,"optional");
	first = 0;
      } else {
	sprintf(params,"%s,optional",params);
      }
    }
    if((desc->paramdesc.wParamFlags & PARAMFLAG_FHASDEFAULT) == PARAMFLAG_FHASDEFAULT) {
      if(first) {
	sprintf(params,"hasdefault");
	first = 0;
      } else {
	sprintf(params,"%s,hasdefault",params);
      }
    }
    sprintf(tmp2,"[%s] %s",params,tmp);
    strcpy(tmp,tmp2);
  }
  return strdup(tmp);
}

SEXP com_object_trace(SEXP handle)
{
  LPDISPATCH dsp = com_getObject (com_getHandle (handle));
  HRESULT hr;
  if (dsp == NULL) {
    RCOM_ERR(printf ("com_object_trace: handle not valid\n"));
    return R_NilValue;
  }
  ITypeInfo* typeinfo = 0;

  hr = dsp->GetTypeInfo(0,LOCALE_USER_DEFAULT,&typeinfo);
  if(FAILED(hr)) {
    RCOM_ERR(printf("com_object_trace: failed %8x GetTypeInfo()\n",hr));
    return R_NilValue;
  }
  TYPEATTR* attr = 0;
  hr = typeinfo->GetTypeAttr(&attr);
  if(FAILED(hr)) {
    RCOM_ERR(printf("com_object_trace: failed %8x getting TYPEATTR\n",hr));
    typeinfo->Release();
  }
  /* trace global type attributes/information */
  RCOM_ERR(printf("COM object information:\n"));
  RCOM_ERR(printf("  type kind: "));
  switch(attr->typekind) {
  case TKIND_ENUM: RCOM_ERR(printf("TKIND_ENUM\n")); break;
  case TKIND_RECORD: RCOM_ERR(printf("TKIND_RECORD\n")); break;
  case TKIND_MODULE: RCOM_ERR(printf("TKIND_MODULE\n")); break;
  case TKIND_INTERFACE: RCOM_ERR(printf("TKIND_INTERFACE\n")); break;
  case TKIND_DISPATCH: RCOM_ERR(printf("TKIND_DISPATCH\n")); break;
  case TKIND_COCLASS: RCOM_ERR(printf("TKIND_COCLASS\n")); break;
  case TKIND_ALIAS: RCOM_ERR(printf("TKIND_ALIAS\n")); break;
  case TKIND_UNION: RCOM_ERR(printf("TKIND_UNION\n")); break;
  case TKIND_MAX: RCOM_ERR(printf("TKIND_MAX\n")); break;
  default: RCOM_ERR(printf("unknown\n")); break;
  }
  RCOM_ERR(printf("  implemented interfaces: %d\n",attr->cImplTypes));
  RCOM_ERR(printf("  number of functions:    %d\n",attr->cFuncs));
  RCOM_ERR(printf("  number of variables:    %d\n",attr->cVars));
  /* trace functions */
  int i=0;
  for(i = 0;i < attr->cFuncs;i++) {
    FUNCDESC* funcdesc = 0;
    //    RCOM_TRACE(printf("  function #%d:\n",i));
    hr = typeinfo->GetFuncDesc(i,&funcdesc);
    if(FAILED(hr)) {
      RCOM_ERR(printf("com_object_trace: error %8x retrieving description for function %d\n",hr,i));
    } else {
      BSTR docbstr = NULL;
      hr = typeinfo->GetDocumentation(funcdesc->memid,NULL,&docbstr,NULL,NULL);
      char* docstr = NULL;
      if(SUCCEEDED(hr) && docbstr) {
	docstr = BSTR2ANSI(docbstr);
	SysFreeString(docbstr);
      } else {
	docstr = NULL;
      }
      {
	int j;
	unsigned int cnt;
	BSTR names[funcdesc->cParams+1];
	for(j = 0;j < funcdesc->cParams+1;j++) names[j] = NULL;
	hr = typeinfo->GetNames(funcdesc->memid,names,funcdesc->cParams+1,&cnt);
	if(FAILED(hr)) {
	  RCOM_ERR(printf("com_object_trace: failed %8x retrieving %d names for member id %d\n",
			    hr,funcdesc->cParams+1,funcdesc->memid));
	} else {
	  char* funcname = BSTR2ANSI(names[0]);
	  RCOM_ERR(printf("  [id(%d)",
			    funcdesc->memid));
	  switch(funcdesc->invkind) {
	  case INVOKE_FUNC: break;
	  case INVOKE_PROPERTYGET: RCOM_ERR(printf(", propget")); break;
	  case INVOKE_PROPERTYPUT: RCOM_ERR(printf(", propput")); break;
	  case INVOKE_PROPERTYPUTREF: RCOM_ERR(printf(", propputref")); break;
	  default: RCOM_ERR(printf(", >>unknown<<\n"));
	  }
	  if(docstr) {
	    RCOM_ERR(printf(", helpstring(\"%s\")",docstr));
	    free(docstr);
	  }
	  char* tmp = get_elem_desc(&funcdesc->elemdescFunc);
	  RCOM_ERR(printf("] %s %s(",tmp,funcname));
	  free(tmp);
	  free(funcname);
	  SysFreeString(names[0]);
	  for(j = 1;j < (int) cnt;j++) {
	    funcname = BSTR2ANSI(names[j]);
	    tmp = get_elem_desc(&funcdesc->lprgelemdescParam[j-1]);
	    RCOM_ERR(printf("%s %s",tmp,funcname));
	    free(tmp);
	    if(j < ((int) cnt) -1) RCOM_ERR(printf(","));
	    free(funcname);
	    SysFreeString(names[j]);
	  }
	  RCOM_ERR(printf(")\n"));
	}
      }
      typeinfo->ReleaseFuncDesc(funcdesc);
    }
  }
  /* trace variables */
  /*  for(i = 0;i < attr->cFuncs;i++) {
      }*/
  typeinfo->ReleaseTypeAttr(attr);
  typeinfo->Release();
  return R_NilValue;
}
SEXP com_object_getinfo(SEXP handle)
{
  int protCnt = 0;
  LPDISPATCH dsp = com_getObject (com_getHandle (handle));
  HRESULT hr;
  if (dsp == NULL) {
    RCOM_ERR(printf ("com_object_trace: handle not valid\n"));
    return R_NilValue;
  }
  ITypeInfo* typeinfo = 0;

  hr = dsp->GetTypeInfo(0,LOCALE_USER_DEFAULT,&typeinfo);
  if(FAILED(hr)) {
    RCOM_ERR(printf("com_object_trace: failed %8x GetTypeInfo()\n",hr));
    return R_NilValue;
  }
  TYPEATTR* attr = 0;
  hr = typeinfo->GetTypeAttr(&attr);
  if(FAILED(hr)) {
    RCOM_ERR(printf("com_object_trace: failed %8x getting TYPEATTR\n",hr));
    typeinfo->Release();
    return R_NilValue;
  }
  /* number of functions to put in array: attr->cFuncs */
  SEXP rc = PROTECT(allocVector(VECSXP,attr->cFuncs));
  SEXP namesattribute = PROTECT(allocVector(STRSXP,attr->cFuncs));
  protCnt++;
  protCnt++;
  /* init the member names: "name", "description", "return", "parameters" */
  SEXP membnamesattribute = PROTECT(allocVector(STRSXP,4));
  protCnt++;
  {
    SEXP a1Sexp = PROTECT(allocString(5)); // "name"
    SEXP a2Sexp = PROTECT(allocString(12)); // "description"
    SEXP a3Sexp = PROTECT(allocString(7)); // "return"
    SEXP a4Sexp = PROTECT(allocString(11)); // "parameters"
    protCnt+=4;
    strcpy(CHAR(a1Sexp),"name");
    strcpy(CHAR(a2Sexp),"description");
    strcpy(CHAR(a3Sexp),"return");
    strcpy(CHAR(a4Sexp),"parameters");
    SET_STRING_ELT(membnamesattribute,0,a1Sexp);
    SET_STRING_ELT(membnamesattribute,1,a2Sexp);
    SET_STRING_ELT(membnamesattribute,2,a3Sexp);
    SET_STRING_ELT(membnamesattribute,3,a4Sexp);
  }

  /* add an entry for each function */
  int i=0;
  for(i = 0;i < attr->cFuncs;i++) {
    FUNCDESC* funcdesc = 0;
    hr = typeinfo->GetFuncDesc(i,&funcdesc);
    if(FAILED(hr)) {
      RCOM_ERR(printf("com_object_trace: error %8x retrieving description for function %d\n",hr,i));
      SEXP strSexp = PROTECT(allocString(4));
      protCnt++;
      strcpy(CHAR(strSexp),"###");
      SEXP vecSexp = PROTECT(allocVector(STRSXP,4));
      protCnt++;
      SET_STRING_ELT(namesattribute,i,strSexp);
      SET_STRING_ELT(vecSexp,0,strSexp);
      SET_STRING_ELT(vecSexp,1,strSexp);
      SET_STRING_ELT(vecSexp,2,strSexp);
      SET_STRING_ELT(vecSexp,3,strSexp);
      setAttrib(vecSexp,R_NamesSymbol,membnamesattribute);
      SET_VECTOR_ELT(rc,i,vecSexp);
    } else {
      BSTR docbstr = NULL;
      hr = typeinfo->GetDocumentation(funcdesc->memid,NULL,&docbstr,NULL,NULL);
      char* docstr = NULL;
      if(SUCCEEDED(hr) && docbstr) {
	docstr = BSTR2ANSI(docbstr);
	SysFreeString(docbstr);
      } else {
	docstr = NULL;
      }
      {
	int j;
	unsigned int cnt;
	BSTR names[funcdesc->cParams+1];
	for(j = 0;j < funcdesc->cParams+1;j++) names[j] = NULL;
	hr = typeinfo->GetNames(funcdesc->memid,names,funcdesc->cParams+1,&cnt);
	if(FAILED(hr)) {
	  RCOM_ERR(printf("com_object_trace: failed %8x retrieving %d names for member id %d\n",
			    hr,funcdesc->cParams+1,funcdesc->memid));
	  SEXP strSexp = PROTECT(allocString(4));
	  protCnt++;
	  strcpy(CHAR(strSexp),"***");
	  SEXP vecSexp = PROTECT(allocVector(STRSXP,4));
	  protCnt++;
	  SET_STRING_ELT(namesattribute,i,strSexp);
	  SET_STRING_ELT(vecSexp,0,strSexp);
	  SET_STRING_ELT(vecSexp,1,strSexp);
	  SET_STRING_ELT(vecSexp,2,strSexp);
	  SET_STRING_ELT(vecSexp,3,strSexp);
	  setAttrib(vecSexp,R_NamesSymbol,membnamesattribute);
	  SET_VECTOR_ELT(rc,i,vecSexp);
	} else {
	  char* funcname = BSTR2ANSI(names[0]);
	  SEXP strSexp = PROTECT(allocString(strlen(funcname) + 1));
	  protCnt++;
	  strcpy(CHAR(strSexp),funcname);
	  SEXP vecSexp = PROTECT(allocVector(STRSXP,4));
	  protCnt++;
	  SET_STRING_ELT(namesattribute,i,strSexp);
	  SET_STRING_ELT(vecSexp,0,strSexp);
	  SET_STRING_ELT(vecSexp,1,strSexp);
	  SET_STRING_ELT(vecSexp,2,strSexp);
	  SET_STRING_ELT(vecSexp,3,strSexp);
	  setAttrib(vecSexp,R_NamesSymbol,membnamesattribute);
	  SET_VECTOR_ELT(rc,i,vecSexp);
	  free(funcname);
	  SysFreeString(names[0]);
	  for(j = 1;j < (int) cnt;j++) {
	    funcname = BSTR2ANSI(names[j]);
	    free(funcname);
	    SysFreeString(names[j]);
	  }
	}
      }
      typeinfo->ReleaseFuncDesc(funcdesc);
    }
  }
  /* trace variables */
  /*  for(i = 0;i < attr->cFuncs;i++) {
      }*/
  typeinfo->ReleaseTypeAttr(attr);
  typeinfo->Release();
  setAttrib(rc,R_NamesSymbol,namesattribute);
  UNPROTECT(protCnt);
  return rc;
}

SEXP com_version_get(SEXP sexp)
{
  SEXP lRc = NULL;
  char* lEnd = NULL;

  lRc = PROTECT(allocVector(REALSXP,1));
  REAL(lRc)[0] = 
    strtod(STRINGIFY(RCOM_VERSION_MAJOR) "." STRINGIFY(RCOM_VERSION_MINOR),
	   &lEnd);
  UNPROTECT(1);
  return lRc;
}

HRESULT STDMETHODCALLTYPE
_IClassFactory_CreateInstance(void* _this,
			      LPUNKNOWN pUnkO,
			      REFIID riid,
			      LPVOID* ppvObj);

/** return the COM object for the internal server itself */
SEXP com_this(SEXP null)
{
  RCOM_OBJHANDLE handle = RCOM_NULLHANDLE;
  HRESULT hr;
  LPDISPATCH object = NULL;
  SEXP sexp = NULL;

  hr = _IClassFactory_CreateInstance(NULL,NULL,
				     IID_IDispatch,
				     (LPVOID*) &object);

  if(SUCCEEDED(hr)) {
    handle = com_addObject (object);
    sexp = com_createSEXP(handle);
    if(sexp == R_NilValue) {
      RCOM_ERR(printf("com_this: invalid SEXP for object, releasing object\n"));
      object->Release();
    }
  }
  return sexp;
}
}
